/*
-------------------------------------------------------------------------------
    This file is a part of the Vst Preset Generator
    Copyright (C) 2007  Francois Mazen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
-------------------------------------------------------------------------------
  Name: maindialog.cpp
  Author: Francois Mazen
  Date: 28/05/07 16:25
  Description: the MainDialog class, the core...
-------------------------------------------------------------------------------
*/

#include <QtGui>

#include "maindialog.h"
#include "vstpresetgenerator.h"
#include "vpgxml.h"
#include "vpgtools.h"

//-----------------------------------------------------------------------------
//MainDialog Class
//-----------------------------------------------------------------------------

//constructor
MainDialog::MainDialog(QWidget *parent) : QDialog(parent)
{
     //Qt objects:
        //buttons:
     btnQuit = new QPushButton(tr("&Quit"));
     btnOpen = new QPushButton(tr("&Import informations\nfrom a preset file"));
     btnOpenList = new QPushButton(tr("&Open a Vst list"));
     btnSaveList = new QPushButton(tr("&Save the Vst list"));
     btnGen = new QPushButton(tr("&Generate a preset file !"));
        //label:
     headerDescription = new QLabel(tr("No header loaded"));
     QLabel *lblNbProg = new QLabel(tr("Number of programs:"));
        //containers:
     QWidget *bottomContainer = new QWidget(this);
     QWidget *topContainer = new QWidget(this);
     QWidget *listContainer = new QWidget(topContainer);
     QWidget *headerContainer = new QWidget(topContainer);
        //groupBox:
     QGroupBox *gbFileType = new QGroupBox(tr("File Type"),topContainer);
        //layout:     
     QVBoxLayout *mainLayout = new QVBoxLayout;
     QGridLayout *bottomLayout = new QGridLayout;//2 rows, 3 columns
     QGridLayout *fileTypeLayout = new QGridLayout;//3 rows, 2 columns
     QHBoxLayout *topLayout = new QHBoxLayout;
     QGridLayout *listLayout = new QGridLayout;//2 rows, 2 columns
     QVBoxLayout *headerLayout = new QVBoxLayout;
       //spinbox:
     spbxBank = new QSpinBox;
       //radio buttons:
     rbtnProgram = new QRadioButton(tr("program preset file (.fxp)"),this);
     rbtnBank = new QRadioButton(tr("bank preset file (.fxb)"),this);
       //Scroll Area
     scrollArea = new QScrollArea(this);
     paramContainer = new QFrame();
     QVBoxLayout *paramLayout = new QVBoxLayout(paramContainer);
     paramContainer->setLayout(paramLayout);
       //List
     vstList = new QListView();
       //Image
     vpgLogo = new const QPixmap(VPG_LOGO);
     QLabel *lblLogo = new QLabel;
     lblLogo->setPixmap(*vpgLogo);
     
     scrollArea->setWidget(paramContainer);
     scrollArea->setWidgetResizable(true);
    
     
     //Fill layout
       //list layout
     listLayout->addWidget(btnOpenList,0,0,1,1);
     listLayout->addWidget(btnSaveList,0,1,1,1);
     listLayout->addWidget(vstList,1,0,1,2);
     listContainer->setLayout(listLayout);
       //header layout
     headerLayout->addWidget(btnOpen);
     headerLayout->addWidget(headerDescription);
     headerLayout->addWidget(lblLogo);
     headerLayout->setAlignment(Qt::AlignHCenter);
     headerContainer->setLayout(headerLayout);
       //groupbox
     fileTypeLayout->addWidget(rbtnProgram,0,0,1,1);
     fileTypeLayout->addWidget(rbtnBank,1,0,1,1);
     fileTypeLayout->addWidget(lblNbProg,2,0);
     fileTypeLayout->addWidget(spbxBank,2,1);
     gbFileType->setLayout(fileTypeLayout);
       //topLayout
     topLayout->addWidget(listContainer);
     topLayout->addWidget(headerContainer);
     topLayout->addWidget(gbFileType);  
     topContainer->setLayout(topLayout);
       //bottom layout   
     bottomLayout->addWidget(scrollArea,0,0,1,3);
     bottomLayout->addWidget(btnGen,1,0);
     bottomLayout->addWidget(btnQuit,1,2);
     bottomContainer->setLayout(bottomLayout);
       //global layout
     mainLayout->addWidget(topContainer);
     mainLayout->addWidget(bottomContainer);
     setLayout(mainLayout);
     
     //init values
     gbFileType->setEnabled(false);
     scrollArea->setEnabled(false);
     btnGen->setEnabled(false);
     rbtnBank->setChecked(true);
     spbxBank->setRange(10,100);
     spbxBank->setSingleStep(10);
     spbxBank->setValue(10);
     
     //vst list:
     vpgXml = new VpgXml(this);
     vpgXml->ReadFile(VST_LIST_FILENAME);
     vstList->setModel(vpgXml);
     
     QString title(tr("Vst Preset Generator V"));
     title += VSTPRESETGENERATORVERSION;
     setWindowTitle(title);
     fxHeader = new FxHeader();
     FxHeaderStruct t_fxheaderstruct;
     t_fxheaderstruct.name = "";
     t_fxheaderstruct.fxID = "";
     t_fxheaderstruct.fxVersion = 0;
     t_fxheaderstruct.nParams = 0;
     ShowLabelHeader(&t_fxheaderstruct);
     
     //Qt connexions
       //buttons:
     connect(btnQuit, SIGNAL(clicked()), this, SLOT(close()));
     connect(btnOpen, SIGNAL(clicked()), this, SLOT(OpenPreset()));
     connect(btnGen, SIGNAL(clicked()), this, SLOT(GeneratePreset()));
     connect(rbtnBank, SIGNAL(toggled(bool)), spbxBank, SLOT(setEnabled(bool)));
     connect(rbtnBank, SIGNAL(toggled(bool)), lblNbProg, SLOT(setEnabled(bool)));
     connect(btnOpenList, SIGNAL(clicked()), this, SLOT(OpenVstList()));
     connect(btnSaveList, SIGNAL(clicked()), this, SLOT(SaveVstList()));
       //headers:
     connect(this, SIGNAL(validHeader(FxHeaderStruct*)), 
             fxHeader, SLOT(receiveParameters(FxHeaderStruct*)));
     connect(this, SIGNAL(validHeader(FxHeaderStruct*)), 
             this, SLOT(ShowLabelHeader(FxHeaderStruct*)));
     connect(vpgXml, SIGNAL(validHeader(FxHeaderStruct*)), 
             fxHeader, SLOT(receiveParameters(FxHeaderStruct*)));
     connect(vpgXml, SIGNAL(validHeader(FxHeaderStruct*)), 
             this, SLOT(ShowLabelHeader(FxHeaderStruct*)));
     connect(fxHeader, SIGNAL(headerIsValid(bool)), 
             gbFileType, SLOT(setEnabled(bool)));
     connect(fxHeader, SIGNAL(headerIsValid(bool)), 
             scrollArea, SLOT(setEnabled(bool)));
     connect(fxHeader, SIGNAL(headerIsValid(bool)), 
             btnGen, SLOT(setEnabled(bool)));
     connect(fxHeader, SIGNAL(headerIsValid(bool)),
             this, SLOT(UpdateParamList()));
      //list interaction 
     connect(vstList, SIGNAL(doubleClicked ( const QModelIndex &)),
             vpgXml, SLOT(loadVst(const QModelIndex &)));
     //Pixmap <-> copyright
     //connect(vpgLogo, SIGNAL(clicked()), this, SLOT(ShowCopyright()));
             
     //Icon
     //setWindowIcon(QIcon(VPG_LOGO));
     
     //for old 640*480 screens:              
     //setFixedSize(600,400);
     
     
             
     
     
}

//-----------------------------------------------------------------------------

void MainDialog::OpenPreset()
{
     QByteArray buffer;
     QString t_fxID;
     qint32 t_fxVersion ;
     qint32 t_nParams ;
     
     int ret;
     
     //get the name of the file
     QString fileName = QFileDialog::getOpenFileName(this, tr("Open Preset"),
                                                "",
                                                tr("Preset Program/Bank (*.fxp *.fxb)"));
     
     //open the file
     QFile file(fileName);
     if (!file.open(QIODevice::ReadOnly))
     {
         //error while opening
         ret = QMessageBox::critical(this, tr("Opening Preset File"),
                   tr("Could not read the file"));                        
         return;
     }
     
     buffer = file.read(4);
     if(buffer.startsWith("CcnK") == false)   //'CcnK'
     {
           //not a preset file
           ret = QMessageBox::critical(this, tr("Opening Preset File"),
                   tr("This is not a valid Vst preset file")); 
           file.close();
           return;
     }
         
     buffer = file.read(4);//chunk size
     buffer = file.read(4);//type of file (bank or program)
     
     //'FxCk' (regular) or 'FPCh' (opaque chunk) Program
     //'FxBk' (regular) or 'FBCh' (opaque chunk) Bank
     if(buffer.startsWith("FxCk") == true)   
     {
              //it's a program or a bank
              buffer = file.read(4);//format version
              
              buffer = file.read(4);//fx unique ID
              t_fxID = buffer.data();
              
              buffer = file.read(4);//fx version
              t_fxVersion = import_long(buffer.data());
              
              buffer = file.read(4);//number of parameters
              t_nParams = import_long(buffer.data());
     }
     else if(buffer.startsWith("FPCh") == true)
     {
          //Opaque chunk data, not suported
          ret = QMessageBox::critical(this, tr("Loading Preset File"),
                   tr("Opaque Chunk method : no supported yet")); 
          return;
     } 
     else if(buffer.startsWith("FxBk") == true)
     {
          //bank program
          
          //jump to the first program and read the parameters
          file.read(4*4 + 128 + 4*3);
          
          buffer = file.read(4);//format version
              
          buffer = file.read(4);//fx unique ID
          t_fxID = buffer.data();
              
          buffer = file.read(4);//fx version
          t_fxVersion = import_long(buffer.data());
              
          buffer = file.read(4);//number of parameters
          t_nParams = import_long(buffer.data());
          
     }
     else if(buffer.startsWith("FBCh") == true)
     {
          //Opaque chunk data, not suported
          ret = QMessageBox::critical(this, tr("Loading Preset File"),
                   tr("Opaque chunk data : no supported yet")); 
          return;
     }    
     else
     {
         //invalid file (no program or no bank)
         ret = QMessageBox::critical(this, tr("Loading Preset File"),
                   tr("This is not a valid Vst preset file")); 
         return;
     }
     
     //Get the name of Vst from the user:
     bool name_ok;
     QString vst_name = QInputDialog::getText(this, tr("Name of the Vst"),
                                          tr("Please enter the name of your Vst:"), QLineEdit::Normal,
                                          tr("My Vst"), &name_ok);
                                          
     FxHeaderStruct *t_fxheaderstruct = new FxHeaderStruct;
     t_fxheaderstruct->fxID = t_fxID;
     t_fxheaderstruct->fxVersion = t_fxVersion;
     t_fxheaderstruct->nParams = t_nParams;  
     t_fxheaderstruct->name = "Invalid name";
     if (name_ok && !vst_name.isEmpty())
     {
         t_fxheaderstruct->name = vst_name;   
     }
     
     //add the new Vst to our list:
     vpgXml->addVst(t_fxheaderstruct);
                   
     //send parmeters to the headers object
     emit validHeader(t_fxheaderstruct);
     
     file.close();
}

//-----------------------------------------------------------------------------

void MainDialog::GeneratePreset()
{
     QString fileName;
     QByteArray buffer;
     int ret;
     qint32 nbPrograms;
     
     bool preset_type_bank = rbtnBank->isChecked();
     
     if(preset_type_bank)
     {
           nbPrograms = spbxBank->value();
     }
     else
     {
         nbPrograms = 1;
     }
     
     //get the name of file
     if(preset_type_bank)
     {
          fileName = QFileDialog::getSaveFileName(this,
                                              tr("Generate Preset Bank"),
                                              "",
                                              tr("Preset Bank (*.fxb)"));
     }
     else
     {
         fileName = QFileDialog::getSaveFileName(this,
                                              tr("Generate Preset Program"),
                                              "",
                                              tr("Preset Program (*.fxp)"));
     }
     
     
     
     
     //open the file
     QFile file(fileName);
     
     if (!file.open(QIODevice::WriteOnly))
     {
         //error while opening
         ret = QMessageBox::critical(this, tr("Saving Preset File"),
                   tr("Could not create the file"));                        
         return;
     }
     
     if(preset_type_bank)
     {
          WriteBank(&file, nbPrograms);
     }
     else
     {
         WriteProgram(&file,nbPrograms);
     }
     
     //Operation success
     QString success(tr("The preset file:\n"));
     success.append(fileName);
     success.append(tr("\nhas been created. Enjoy !"));
     ret = QMessageBox::information(this, tr("Saving Preset File"), success);
     
     file.close();
     
}

//-----------------------------------------------------------------------------

void MainDialog::ShowLabelHeader(FxHeaderStruct* data)
{
     
     QString temp1;
     QString temp2;
     temp1 += tr("Vst name: ");
     temp1 += data->name;
     temp1 += tr("\nVst ID: ");
     temp1 += data->fxID;
     temp1 += tr("\nVst Version: ");
     temp2.setNum(data->fxVersion);
     temp1 += temp2;
     temp1 += tr("\nnumber of parameters: ");
     temp2.setNum(data->nParams);
     temp1 += temp2;
     
     headerDescription->setText(temp1);
     
}

//-----------------------------------------------------------------------------
void MainDialog::WriteProgram(QFile *file, qint32 no_program)
{
     QByteArray buffer;
     qint32 val, no_param;
     float valf, min, max;
     
     //cMagic
     buffer = "CcnK";
     file->write(buffer);
     
     //size of byte * number of parameters
     val = 4 * 5 + 28 + 4 * fxHeader->Data.nParams; 
     
     buffer.clear();
     buffer.resize(4);
     export_long(val,buffer.data());
     file->write(buffer);
     
     ///< 'FxCk' (regular) or 'FPCh' (opaque chunk)
     buffer = "FxCk";
     file->write(buffer);
     
	 ///< format version (currently 1)
	 val = 1;
	 buffer.clear();
     buffer.resize(4);
     export_long(val,buffer.data());
     file->write(buffer);
	 
	 ///< fx unique ID
	 buffer.clear();
	 buffer += fxHeader->Data.fxID;
	 file->write(buffer);
	 
	 ///< fx version
	 buffer.clear();
     buffer.resize(4);
     export_long(fxHeader->Data.fxVersion,buffer.data());
	 file->write(buffer);
     
     buffer.clear();
     buffer.resize(4);
     //< number of parameters
     export_long(fxHeader->Data.nParams,buffer.data());
     file->write(buffer);
	 
     ///< program name (null-terminated ASCII string) !!!![28]!!!!
     buffer.clear();
     buffer.setNum(no_program);
     buffer.prepend("Vst Preset Generator ");
     file->write(buffer,28);
     
     //Float parameters
        
     for(no_param = 0; no_param < fxHeader->Data.nParams; no_param ++)
     {
          buffer.clear();
          buffer.resize(4);
          min = paramList[no_param]->getMin();
          max = paramList[no_param]->getMax();
          valf = (((float) qrand()) / ((float)RAND_MAX)) * (max - min) + min;
          char* pt = (char*)&valf; 
          
          //inversion because little endian notation
          buffer[3] = pt[0];
          buffer[2] = pt[1];
          buffer[1] = pt[2];
          buffer[0] = pt[3];
          
          file->write(buffer);
          
     }//params
        
     
}

//-----------------------------------------------------------------------------

void MainDialog::WriteBank(QFile *file, qint32 nb_programs)
{
     QByteArray buffer;
     qint32 val, no_program;
     
     //cMagic
     buffer = "CcnK";
     file->write(buffer);
     
     //size of byte * number of program
     val = 4 * 5 + 128 + nb_programs * (4 * 5 + 28 + 4 * fxHeader->Data.nParams);
     buffer.clear();
     buffer.resize(4);
     export_long(val,buffer.data());
     file->write(buffer);
     
     ///< 'FxCk' (regular) or 'FPCh' (opaque chunk)
     buffer = "FxBk";
     file->write(buffer);
     
	 ///< format version (currently 1)
	 val = 1;
	 buffer.clear();
     buffer.resize(4);
     export_long(val,buffer.data());
     file->write(buffer);
	 
	 ///< fx unique ID
	 buffer.clear();
	 buffer += fxHeader->Data.fxID;
	 file->write(buffer);
	 
	 ///< fx version
	 buffer.clear();
     buffer.resize(4);
     export_long(fxHeader->Data.fxVersion,buffer.data());
	 file->write(buffer);
     
     buffer.clear();
     buffer.resize(4);
     //number of programs
     export_long(nb_programs,buffer.data());
     file->write(buffer);
	 
	 //future space for bank preset
     buffer.clear();
     file->write(buffer, 128);
	 
	 //Write all programs
	 QTime midnight(0, 0, 0);//dummy QTime object
     qsrand(midnight.secsTo(QTime::currentTime()));
	 for(no_program = 0; no_program < nb_programs; no_program++)
	 {
         WriteProgram(file, no_program);
     }//programs
     
}

//-----------------------------------------------------------------------------
void MainDialog::UpdateParamList()
{
     int i;
     ParamWidget *temp;
     
     //delete the old widgets
     if(paramList.isEmpty() == false)
     {
            for(i=0;i < paramList.size(); i++)
            {
               paramContainer->layout()->removeWidget(paramList[i]);
               delete(paramList[i]);
            }
     }     
     
     paramList.clear();
     //paramList.resize(fxHeader->nParams);
     
     for(i =1; i<=fxHeader->Data.nParams; i++)
     {
           temp = new ParamWidget(i);
           paramList.append(temp);
           paramContainer->layout()->addWidget(temp);
     }
}

//-----------------------------------------------------------------------------
void MainDialog::SaveVstList()
{
     //get the name of the file
     QString fileName = QFileDialog::getSaveFileName(this, tr("Save Vst List"),
                                                "",
                                                tr("Vst Preset Generator file (*.vpg)"));
          
     vpgXml->WriteFile(qPrintable(fileName));
}

//-----------------------------------------------------------------------------
void MainDialog::OpenVstList()
{
     //get the name of the file
     QString fileName = QFileDialog::getOpenFileName(this, tr("Open Vst List"),
                                                "",
                                                tr("Vst Preset Generator file (*.vpg)"));
     
     vpgXml->ReadFile(qPrintable(fileName));
     
     //emit an empty header
     FxHeaderStruct *invalid = new FxHeaderStruct;
     invalid->name = "";
     invalid->fxID = "";
     invalid->fxVersion = 0;
     invalid->nParams = 0;
     
     emit validHeader(invalid);
}

//-----------------------------------------------------------------------------
void MainDialog::ShowCopyright()
{
     //TODO: show copyright information (GNU GPL)
     /*int ret = QMessageBox::information(this, tr("VST Preset Generator"),
                   tr("The document has been modified.\n"
                      "Do you want to save your changes?"));*/
}

